import pytest
import numpy as np
import pandas as pd
from inspect import signature

from ..fakeflares import (flare_model_mendoza2022, flare_model_davenport2014,
                          generate_fake_flare_distribution,
                          merge_fake_and_recovered_events,
                          mod_random,)


from .test_flarelc import mock_flc


def test_mod_random():
    assert mod_random(1, d=True)[0] == pytest.approx(0.48661046)
    assert mod_random(1)[0] != mod_random(1)[0]


def test_merge_fake_and_recovered_events():

    flares = {'istart' : [5, 30],
              'istop' : [15, 36],
              'cstart' : [8, 33],
              'cstop': [18, 39],
              'tstart' : [2335.5846, 2335.5846 + 12.5/24.],
              'tstop' : [2335.5846 + 5./24., 2335.5846 + 15.5/24.],
              'ed_rec' : [650, 250],
              'ed_rec_err' : [46, 15]}
    fakes = {'duration_d' : np.linspace(0.01,1,20),
             'amplitude' : np.linspace(0.01,1,20),
             'ed_inj' : np.linspace(100,50000,20),
             'peak_time' : np.linspace(2320,2340,20),}
    flares = pd.DataFrame(flares)
    fakes = pd.DataFrame(fakes)
    tab = merge_fake_and_recovered_events(fakes, flares)
    assert tab.size == 240
    assert tab.shape[0] == 20
    match = tab.dropna(how='any')
    assert match.shape[0] == 1
    row = match.iloc[0]
    assert row.peak_time >= row.tstart
    assert row.peak_time <= row.tstop
    assert row.ed_inj > row.ed_rec

def test_generate_fake_flare_distribution():

    n = 500

    dur, ampl = generate_fake_flare_distribution(n)
    assert (dur <= 20).all()
    assert (dur >= 1e-4).all()
    assert (ampl <= 1e2).all()
    assert (ampl >= 1e-4).all()
    assert len(dur) == n
    assert len(ampl) == n

    dur, ampl = generate_fake_flare_distribution(n, ampl=[1e-4, 5], dur=[0.005, 0.012],  mode='uniform')
    assert (dur <= 0.012).all()
    assert (dur >= 0.0045).all()
    assert (ampl <= 5).all()
    assert (ampl >= 1e-4).all()
    assert len(dur) == n
    assert len(ampl) == n

  

# def test_inject_fake_flares():
#     flc = mock_flc(detrended=True)
#     np.random.seed(84712)
#     flc = flc.find_gaps()
#     fake_flc = inject_fake_flares(flc)
#
#     assert fake_flc.fake_flares.size == 20
#
#     columns = set(fake_flc.fake_flares.columns.values.tolist())
#     test_columns = {'amplitude', 'duration_d', 'ed_inj', 'peak_time'}
#     assert len(columns & test_columns) == 4
#     assert fake_flc.detrended_flux_err.all() >= 1e-10
#     assert fake_flc.detrended_flux.all() <= 1.
#     assert fake_flc.detrended_flux.shape == flc.detrended_flux.shape
#     flc = mock_flc(detrended=False)
#     np.random.seed(84712)
#     flc = flc.find_gaps()
#     fake_flc = inject_fake_flares(flc, inject_before_detrending=True)
#
#     assert fake_flc.fake_flares.size == 20
#     columns = set(fake_flc.fake_flares.columns.values.tolist())
#     test_columns = {'amplitude', 'duration_d', 'ed_inj', 'peak_time'}
#     assert len(columns & test_columns) == 4
#     assert fake_flc.flux_err.all() >= 1e-10
#     assert fake_flc.flux.all() <= 1.
#     assert fake_flc.flux.shape == flc.flux.shape

def test_flare_model_davenport2014_and_equivalent_duration():

    n = 1000
    time = np.arange(0, n/48, 1./48.)
    x = time * 60.0 * 60.0 * 24.0

    # Test a large flare without upsampling
    fl_flux = flare_model_davenport2014(time, 11.400134, 1.415039, 110.981950)
    integral = np.sum(np.diff(x) * fl_flux[:-1])
    assert integral == pytest.approx(1.22e7,rel=1e-2)
    
    # Test a flare with 0 amplitude
    fl_flux = flare_model_davenport2014(time, 11.400134, 1.415039, 0)
    integral = np.sum(np.diff(x) * fl_flux[:-1])
    assert integral == 0.

    # test a large flare with upsampling
    fl_flux = flare_model_davenport2014(time, 11.400134, 1.415039, 110.981950, upsample=True)
    integral = np.sum(np.diff(x) * fl_flux[:-1])
    assert integral == pytest.approx(1.22e7,rel=1e-2)
    
    
    # Test a smaller undersampled flare
    fl_flux = flare_model_davenport2014(time, 11.400134, 1/48., 1.0)
    x = time * 60.0 * 60.0 * 24.0
    integral = np.sum(np.diff(x) * fl_flux[:-1])
    assert integral == pytest.approx(1453.1179,rel=1e-2)
    
    # Test the amplitude
    fl_flux = flare_model_davenport2014(time, 1.734, 15, 1.0)
    assert np.max(fl_flux) == pytest.approx(1,rel=1e-2)

def test_flare_model_mendoza2022():

    n = 1000
    time = np.arange(-2, n/48, 1./48.)
    x = time * 60.0 * 60.0 * 24.0

    # Test a large flare without upsampling
    fl_fluxu = flare_model_mendoza2022(time, 11.400134, 1.415039, 110.981950)
    integralu = np.sum(np.diff(x) * fl_fluxu[:-1])
    assert integralu == pytest.approx(2.26e7,rel=1e-2)
    
    # Test a flare with 0 amplitude
    fl_flux0 = flare_model_mendoza2022(time, 11.400134, 1.415039, 0)
    integral0 = np.sum(np.diff(x) * fl_flux0[:-1])
    assert integral0 == 0.

    # test a large flare with upsampling
    fl_fluxup = flare_model_mendoza2022(time, 10., 1.5, 100, upsample=True,uptime=10)
    integralup = np.sum(np.diff(x[:-1]) * fl_fluxup[:-2])
    assert integralup == pytest.approx(2.2e7,rel=1e-2)
    
    
    # Test a smaller undersampled flare
    fl_fluxus = flare_model_mendoza2022(time, 10, .25, 1.0)
    integralus = np.sum(np.diff(x) * fl_fluxus[:-1])
    assert integralus == pytest.approx(44055.6396,rel=1e-2)
    
    # Test the amplitude
    fl_flux = flare_model_mendoza2022(time, 2, 15, 1)
    assert np.max(fl_flux) == pytest.approx(1,rel=10e-2)
    fl_fluxa = flare_model_mendoza2022(time, 2, 15, 1)




import pytest
import numpy as np

from ..utils import generate_random_power_law_distribution
from ..wheatland import (occurrence_probability_posterior,
                         uninformative_prior,
                         flaring_rate_likelihood,
                         logit,
                         BayesianFlaringAnalysis,
                         calculate_joint_posterior_distribution,
                         generate_fake_data,
                         )


def test_logit():
    '''Test logit function by logifying
    some random function'''

    def func(x,a,b):
        return x / (b - a)
    log_uninf_prior = logit(func)
    up = func(3,2,6)
    logup = log_uninf_prior(3,2,6)
    assert np.log(0.75) == logup


def test_flaring_rate_likelihood():
    '''Test flaring rate likelihood with a
       5 flares per day rate.
    '''
    # Find the likelihood

    rates = np.linspace(1e-1, 20, 10000)
    posterior = flaring_rate_likelihood(rates, 75, 15., norm=True)

    #-----------------------------------------

    # Check some values:
    assert np.sum(posterior) == pytest.approx(1.)
    assert rates[np.argmax(posterior)] == pytest.approx(5., rel=1e-3)
    assert rates[np.argmin(np.abs(np.max(posterior)/2.-posterior))] ==  pytest.approx(4.351065, rel=1e-4)


def test_uninformative_prior():
    '''Test some exceptions that should be mostly ignored.'''

    # Define test values:
    vals = [(1,2,3), (2,1,3), (np.nan, 4, 10), (3, 2, 2)]

    # Run prior calculation on values:
    res = []
    for rate, minrate, maxrate in vals:
        res.append(uninformative_prior(rate, minrate, maxrate))

    # Check results:
    assert res == [0,0.5,0,0]


def test_occurrence_probability_posterior():
    '''Test the posterior using a uniform prior'''

    # Simulate the results from Wheatland 2004 in Fig. 1:
    # Use their precise values:

    t = 5 #total observation time in days
    cadence = 4 #observations per hour
    obstimes = np.linspace(3000,3000+t,t*24*4) # 15 min cadence observations
    flaresperday = 5. # average flaring rate in flares per day
    np.random.seed(3000)
    times = obstimes[np.where(np.random.poisson(lam=1. / 24. / cadence * flaresperday, size=t*24*4))[0]]
    size = len(times)
    events = generate_random_power_law_distribution(1, 100000, -.8, size=size, seed=778)
    #time interval to consider for prediction
    Tprime = 5. # if bayesian blocks used: bins[-1] - bins[-2]
    mined = 100 # min ED value we want to predict a rate for, same as S2 in Wheatland paper
    deltaT = 1. # predict rate of flares above threshold for deltaT days in the futures
    alpha = 1.8 # fix power law exponent for now
    threshed = 1. # detection sensitivity limit
    # number of observations
    Mprime = size# if bayesian blocks used: values[-1]

    # Find the posterior distribution:

    x = np.linspace(1e-8,1-1e-8,10000)
    predicted_distr = occurrence_probability_posterior(x, alpha, mined, Tprime,
                                                       Mprime, deltaT, threshed)

    #--------------------------------------------------

    # Check some values:
    # TODO: use more restrictive check because you know how to seed numpy random generators now.

    assert x[np.argmax(predicted_distr)] > 0.110
    assert x[np.argmax(predicted_distr)] < 0.131
    assert x[np.argmin(np.abs(np.max(predicted_distr)/2.-predicted_distr))] ==  pytest.approx( 0.12741274, rel=1e-4)

    # --------------------------------------------------

    # If debugging is needed later:
    # fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,4))
    # ax.plot(x, predicted_distr, color="r",
    #         label='ED={} s,  t={} d'.format(mined, deltaT))

    # ax.set_xlabel(r'$\varepsilon$: probability of one event'+'\nabove ED within t after last observation')
    # ax.set_ylabel(r'normalized $P_{\varepsilon}(\varepsilon)$')
    # ax.set_xlim(0,0.25)
    # ax.legend()


def test_sample_posterior_with_mcmc():
    '''Integration test using the example
    in Wheatland (2004) and our own
    likelihoods and priors.
    '''

    # Create some fake data:

    t = 5 #total observation time in days, must be int
    cadence = 4 #observations per hour
    obstimes = np.linspace(3000,3000+t,t*24*cadence) # 15 min cadence observations
    flaresperday = 10. # average flaring rate in flares per day
    times = obstimes[np.where(np.random.poisson(lam=1. / 24. / cadence * flaresperday, size=t * 24 * cadence))[0]]
    size = len(times)
    alpha_prior = 1.8 # fix power law exponent for now
    events = generate_random_power_law_distribution(1, 1000, -alpha_prior + 1., size=size, seed=788)
    #time interval to consider for prediction
    Tprime = 5#np.max(times) -  np.min(times)# if bayesian blocks used: bins[-1] - bins[-2]
    mined = 100 # min ED value we want to predict a rate for, same as S2 in Wheatland paper
    deltaT = 1.# predict rate of flares above threshold for deltaT days in the futures
    threshed = 1 # detection sensitivity limit
    Mprime = size# if bayesian blocks used: values[-1]
    rate_prior = flaresperday / np.abs(alpha_prior - 1.) * np.power(mined, -alpha_prior +1.) # evaluate cumulative FFD fit at mined
    eps_prior = 1 - np.exp(-rate_prior * deltaT) # calculate poisson probability.  Eqn (5) in Wheatland (2004)

    #---------------------------------

    # Define your log-likelihood function:

    def loglikelihood(theta, *args):
        def prior(x):
            return uninformative_prior(x, 1.25, 2.25)
        return calculate_joint_posterior_distribution(theta, *args, prior)

    # Create a BayesianAnalysisObject

    BFA = BayesianFlaringAnalysis(mined=mined, Tprime=Tprime, deltaT=deltaT, alpha_prior=alpha_prior, eps_prior=eps_prior,
                              threshed=threshed, Mprime=Mprime, events=events, loglikelihood=loglikelihood)

    # Run MCMC to sample the posterior distribution

    BFA.sample_posterior_with_mcmc()

    #---------------------------------

    # Check that the function ran through

    assert BFA.samples.shape == (31800, 2) #this goes wrong if default values for steps/cutoff/nwalkers are changed
    assert BFA.samples[:,1].max() < 2.25 # as defined by prior on alpha
    assert BFA.samples[:,1].min() > 1.25 # as defined by prior on alpha
    assert BFA.samples[:,0].min() > 0 # beta is always positive

    # Check that the results are correct maybe?


def test_calculate_percentiles():
    '''Test if the percentiles come out
    correctly in a random gauss shaped sample.
    '''
    # Create some data with a seed:

    np.random.seed(seed=590)
    samples = np.array([np.random.normal(loc=1.0, scale=.2, size=1000),
                        np.random.normal(loc=2.0, scale=.4, size=1000)]).T
    BFA = BayesianFlaringAnalysis(samples=samples)

    #-----------------------------------

    # Run the function:

    pct = BFA.calculate_percentiles()

    #-----------------------------------

    # Check the results:

    assert (np.round(pct[0],6) == np.round((1.008030476417989, 0.19516752006028892, 0.21571271977738948), 6)).all()
    assert (np.round(pct[1],6) == np.round((1.980568570082061, 0.4199525752578359, 0.37256627621632066), 6)).all()


def test_generate_fake_data():
    
    # ---------------------------------------------------------
    # Integration test
    fake = generate_fake_data(13, 120, 5, 2., seed=9000)
    assert fake["deltaT"] == 13. 
    assert fake["deltaT"] == fake["Tprime"]
    assert fake["alpha_prior"] ==  2.
    assert fake["beta_prior"] ==  5.
    assert fake["threshed"] == 1. 
    assert fake["mined"] ==  pytest.approx(78.78287887 * 10., rel=.01)
    assert fake["beta_prior"] == 5. 
    assert fake["eps_prior"] == pytest.approx(0.07919338333736559)
    print(len(fake["events"]))
    assert len(fake["events"]) == pytest.approx(65,abs=10) # 65 would be the perfect as in 5x13
    assert (fake["events"]>1.).all()

    # -----------------------------------------------------------
    # In case of too high flaring rate relative to time sampling:
    with pytest.raises(ValueError):
        fake = generate_fake_data(13, 120, 1000, 2., seed=9000)



import pytest


import numpy as np
import pandas as pd

from ..flarelc import FlareLightCurve

from ..customdetrend import (custom_detrending,
                       estimate_detrended_noise,
                       remove_exponential_fringes,
                       remove_sines_iteratively,
                       measure_flare)

from ..altai import find_iterative_median

cases = [(.05, 0.005, 1.5, 24.4, 1.5, 0.1),
         (.1, 0.005, 1.5, 14.4, 1.5, 0.5),
         (.1, 0.05, 1.5, 8, 1.5, 0.5),
         (.01, .1, .5, 1, -.5, 1.5),
         (.3, .05, .5, 30, -.5, 0.25),
         (.01, .1, .5, 2.2, -.1, .1),
         ]

@pytest.mark.parametrize("a1,a2,period1,period2,quad,cube", cases)
def test_custom_detrending(a1, a2, period1, period2, quad, cube,):
    
    # fix uncertainty
    errorval = 15.
    np.random.seed(40)
    lc = generate_lightcurve(errorval, a1, a2, period1, period2, quad, cube)

#     lc.plot()
    flcc = custom_detrending(lc)

    flccc = estimate_detrended_noise(flcc, mask_pos_outliers_sigma=2.5, 
                                     std_window=100)

    flccc = find_iterative_median(flccc)


    flares = flccc.find_flares(addtail=True).flares
    print(flares.ed_rec)
    print(flares.ampl_rec)

    # check that uncertainty is 
    assert np.nanmedian(flccc.detrended_flux_err) == pytest.approx(errorval, abs=2)


    compare = pd.DataFrame({'istart': {0: 5280, 1: 13160, 2: 23160},
                            'istop': {0: 5346, 1: 13163, 2: 23175}})
        
    assert (flares[["istart"]] == compare[["istart"]]).all().all()

    edrec = np.array([802.25, 4.7907, 40.325])
    amplrec = np.array([0.28757, 0.03004, 0.064365])

    for i in range(3):
        assert [float(flares[["istop"]].values[i])] == pytest.approx([float(compare[["istop"]].values[i])],
                                                            rel=.09)
    
    for i in range(3):
        assert [flares.ed_rec.values[i]] == pytest.approx([edrec[i]], rel=0.2)

    for i in range(3):
        assert flares.ampl_rec.values[i] == pytest.approx(amplrec[i], rel=0.25)

    return 




cases = [(0.1, 3., 0.1, 5.),
         (0.1, 3., 0.1, 10.),
         (0.1, 8., 0.1, 10.),
         (0.05, 4., 0.1, 10.),
         (0.05, 4., 0.1, 3.),]

@pytest.mark.parametrize("a,b,c,d", cases)
def test_remove_sines_iteratively(a, b, c, d):

    # define light curve with two sinusoidal modulation
    x = np.linspace(10, 40, 1200)
    y1 = 20. + np.random.normal(0, .01, 1200) + a * np.sin(b * x) + c * np.sin(d * x)
    
    flc = FlareLightCurve(time=x, flux=y1, flux_err=np.full_like(y1, .01),)
    flc.detrended_flux = y1
    flc.detrended_flux_err = np.full_like(y1, .01)

    # find median
    flc = find_iterative_median(flc)
#     flc.plot()
    
    # apply function
    flcd = remove_sines_iteratively(flc)

#     plt.plot(flcd.time, flcd.flux)
#     plt.plot(flcd.time, flcd.detrended_flux)
    
    # do some checks
    assert flcd.detrended_flux.std() == pytest.approx(0.01, rel=1e-1)
    assert flcd.detrended_flux.max() < 20.2
    assert flcd.detrended_flux.min() > 19.8

cases = [(1., 40.,20.,1.,10.),
         (1., 40.,20.,-1.,10.),
         (-1., 40.,20.,-1.,10.),
         (-1., 40.,20.,1.,10.),
         (1., 40.,5.,1.,10.),
         ]

@pytest.mark.parametrize("a,b,median,c,d", cases)
def test_remove_exponential_fringes(a,b,median,c,d):
    
    # seed numpy random to exclude outliers
    np.random.seed(42)
    
    # define light curve with two positive fringes
    x = np.linspace(10,40,1200)
    y1 = (a*np.exp(-1 * (b - x) * 2) +
          median +
          c*np.exp((d - x) * 2) +
          np.random.normal(0, .0005*median, 1200))
    y1[800:810] = median + median * .05 * np.linspace(1,0,10)
    
    

    # define lightcurve
    flc = FlareLightCurve(time=x, flux=y1, flux_err=np.full_like(y1, .0005*median))
    flc.detrended_flux = y1
    flc.detrended_flux_err = np.full_like(y1, .0005*median)
    
    # get iterative median
    flc = find_iterative_median(flc)
    
    # run the function
    flcd = remove_exponential_fringes(flc)

#     plt.plot(flcd.time, flcd.flux)
#     plt.plot(flcd.time, flcd.detrended_flux)

    # do some checks

#     print(flcd.detrended_flux.std(), flcd.detrended_flux.min(), flcd.detrended_flux.max())
    assert flcd.detrended_flux[:799].std() == pytest.approx(.0005*median, rel=1e-1)
    assert flcd.detrended_flux.max() == pytest.approx(median * 1.05)
    assert flcd.detrended_flux.min() > median * 0.995

def generate_lightcurve(errorval,  a1, a2,period1, period2, quad, cube,
                        mean=3400.):
    
    """Generate wild light curves with variability on several
    timescales.
    
    Returns:
    ---------
    FlareLightCurve with time, flux, and flux_err attributes
    """
    time = np.arange(10, 10 + 10 * np.pi,.0008)

    # define the flux
    flux = (np.random.normal(0,errorval,time.shape[0]) +
            mean + 
            a1*mean*np.sin(period1*time +1.)  +
            a2*mean*np.sin(period2*time) +
            quad*(time-25)**2 -
            cube*(time-25)**3)

    # add a gap in the data
    flux[5600:7720] = np.nan

    # add big and long flare
    l = 66
    flux[5280:5280 + l] = flux[5280:5280 + l] + np.linspace(1000,250,l)

    # add tiny flare
    l = 3
    flux[15280:15280 + l] = flux[15280:15280 + l] + np.linspace(100,60,l)

    # add intermediate flare
    l, s = 15, 25280
    flux[s:s + l] = flux[s:s + l] + np.linspace(200,60,l)

    # typically Kepler and TESS underestimate the real noise
    err = np.full_like(time,errorval/3*2)

    # define FLC
    return FlareLightCurve(time=time, flux=flux, flux_err=err)



def test_estimate_detrended_noise():
    
    # setup light curve
    time = np.linspace(10,30,200)
    
    # seed numpy to get the same error array
    np.random.seed(30)
    
    # define flux with gaussian noise and baseline flux
    flux = np.random.normal(0,40, time.shape[0]) + 200.
    
    # define light curve
    flc = FlareLightCurve(time=time)
    flc.detrended_flux=flux

    # this should work
    flces = estimate_detrended_noise(flc, mask_pos_outliers_sigma=2.5, 
                                 std_window=100)
    
    # error should be similar to input error of 40
    np.median(flces.detrended_flux_err.value) == pytest.approx(41.38048677022836)

    # re-seed and add a flare
    np.random.seed(30)
    flux = np.random.normal(0,40, time.shape[0]) + 200.
    flux[120:124] = [500,380,300,270]
    flc = FlareLightCurve(time=time)
    flc.detrended_flux = flux

    # should mask flare, error should not grow
    flces = estimate_detrended_noise(flc, mask_pos_outliers_sigma=2.5, 
                                 std_window=100)

    np.median(flces.detrended_flux_err.value) == pytest.approx(41.24232394552432)

    # re-seed and add some NaNs
    np.random.seed(30)
    flux = np.random.normal(0,40, time.shape[0]) + 200.
    flux[120:124] = [500,380,300,270]
    flux[30:40] = np.nan
    flc = FlareLightCurve(time=time) 
    flc.detrended_flux = flux

    # should work regardless
    flces = estimate_detrended_noise(flc, mask_pos_outliers_sigma=2.5, 
                                     std_window=100)

    # error should not change too much
    np.median(flces.detrended_flux_err.value) == pytest.approx(41.23144256208637)
    
    
    
def test_measure_flare():
    """Simple test: Generate light curve with flare,
    detrend, and manually measure the flare.
    """
    # generate LC
    flc = generate_lightcurve(15,.01,.03,4,.3,.1,.02)
    
    # de-trend LC
    flcc = custom_detrending(flc)

    # measure flare
    measure_flare(flcc,5280,5280+66)

    # get pandas.Series
    measured_flare = flcc.flares.iloc[0]

    # do checks
    assert measured_flare.istart == 5280
    assert measured_flare.istop == 5346
    assert measured_flare.tstart == pytest.approx(14.224)
    assert measured_flare.tstop ==  pytest.approx(14.276800)
    assert measured_flare.ed_rec == pytest.approx((250 + 750 * 0.5) / 3400 * 0.052800 * 24 * 3600, rel=.03)
    assert measured_flare.tstop -measured_flare.tstart == measured_flare.dur


import pytest

import numpy as np

from ..lcio import from_path, from_mast, _from_path_AltaiPony
from . import pathkepler, pathtess



@pytest.mark.parametrize("path,mode,ID,mission,campaign,quarter,sector",
                         [(pathkepler,"LC", 10002792, "Kepler", None, 2, None ),
                          (pathtess,"LC", 358108509, "TESS", None, None, 1)
                          ])

def test_from_path(path, mode, ID, mission, campaign, quarter, sector):
    
    flc = from_path(path, mode=mode, mission=mission)
    assert flc.targetid == ID
    assert flc.mission == mission
    assert flc.flux.value.shape[0] == flc.time.value.shape[0]
    if campaign is not None:
        assert flc.campaign == campaign
    if quarter is not None:
        assert flc.quarter == quarter
    if sector is not None:    
        assert flc.sector == sector    
    assert (np.isnan(flc.flux.value) == False).all()
    assert (np.isnan(flc.time.value) == False).all()
    
    assert np.isnan(flc.detrended_flux.value).all()
    assert (~np.isnan(flc.flux_err.value)).all()
    assert flc.flares.shape[0] == 0
    flct = flc[10:20]
    assert flct.flux.value.shape[0] == flct.time.value.shape[0]
    assert flct.flux.value.shape[0] == flct.flux_err.value.shape[0]
    assert flct.flux.value.shape[0] == flct.detrended_flux_err.value.shape[0]
    assert flct.flux.value.shape[0] == 10
    
    
def test__from_path_AltaiPony():
    flc = from_path("altaipony/examples/kplr010002792-2010174085026_llc.fits","LC", "Kepler")
    path = "altaipony/examples/pony010002792-2010174085026_llc_test_from_path_AltaiPony.fits"
    #flc = from_path(pathkepler, mode="LC", mission="Kepler")
    flc.to_fits(path, overwrite=True)
    rflc = _from_path_AltaiPony(path)
    assert rflc.channel == flc.channel
    assert rflc.quarter == flc.quarter
    assert rflc.ra == flc.ra
    assert rflc.dec == flc.dec
    assert rflc.mission == "Kepler"
    kws = ['time', 'flux', 'flux_err', 'centroid_col',
           'centroid_row', 'quality', 'cadenceno',
           'detrended_flux', 'detrended_flux_err',]
    assert (len(rflc.flux.value) == np.array([len(getattr(rflc,x)) for x in kws])).all()
    assert rflc.targetid == 10002792
    assert rflc.origin == "FLC"
    assert np.isnan(rflc.detrended_flux.value).all()
    assert np.isnan(rflc.detrended_flux_err.value).all()
    
    

@pytest.mark.parametrize("ID,mission,c,mode,cadence,sector,campaign,quarter,lflc",
                         [("TIC 395130640","TESS", 11,"LC", "short", 11, None, None,1 ),
                          ("KIC 9726699", "Kepler", 6, "LC", "long", None, None, 6,1),
                          ("KIC 100004076", "Kepler", 14, "LC", "short", None, None, 14,3),
                          ("TIC 395130640","TESS", None,"LC", "short", 11, None, None,2),
                          ("KIC 9726699", "Kepler", None, "LC", "long", None, None, 0,15),
                          ("KIC 100004076", "Kepler", None, "LC", "short", None, None, 14,3)
                          ])
def test_from_mast(ID, mission, c, mode, cadence, sector, campaign, quarter, lflc):
    flc = from_mast(ID, mission, c, mode=mode, cadence=cadence)
    # Only for the KIC 100004706 target:
    if isinstance(flc, list):
        assert len(flc) >= lflc 
        flc = flc[0]
    # -----------------------------------    
    assert flc.targetid == int(ID.split(" ")[1])
    assert flc.mission == mission
 
    assert flc.flux.value.shape[0] == flc.time.value.shape[0]
    if campaign is not None:
        assert flc.campaign == campaign
    if quarter is not None:
        assert flc.quarter == quarter
    if sector is not None:    
        assert flc.sector == sector
    assert (np.isnan(flc.flux.value) == False).all()
    assert (np.isnan(flc.time.value) == False).all()
    print(flc.columns)
    assert np.isnan(flc.detrended_flux.value).all()
    assert (~np.isnan(flc.flux_err.value)).all()
    assert flc.flares.shape[0] == 0
    flct = flc[10:20]
    assert flct.flux.value.shape[0] == flct.time.value.shape[0]
    assert flct.flux.value.shape[0] == flct.flux_err.value.shape[0]
    assert flct.flux.value.shape[0] == flct.detrended_flux_err.value.shape[0]
    assert flct.flux.value.shape[0] == 10
    




import pytest

import numpy as np

from ..flarelc import FlareLightCurve
from ..utils import k2sc_quality_cuts, split_gaps, expand_mask

import copy

def test_expand_mask():
    # set up a test array with a single outlier and a sequence
    a = np.ones(25).astype(int)
    a[3] = 0
    a[[15,16]] = 0

    # check result
    assert (expand_mask(copy.deepcopy(a), longdecay=2) == 
            np.array([1, 1, 1, 0, 1,
                      1, 1, 1, 1, 1,
                      1, 1, 1, 1, 0,
                      0, 0, 0, 0, 1,
                      1, 1, 1, 1, 1])).all()


def test_k2sc_quality_cuts():
    a = np.array([1, 3., 5, 98.])
    b = np.array([2, 4., np.nan, 20])
    c = np.array([3, np.inf , 20.1, 5])
    d = np.array([3, 20. , 20.1, 5])
    data = FlareLightCurve(time=a, flux=d)
    data["pos_corr1"] = b
    data["pos_corr2"] = c
    data2 = k2sc_quality_cuts(data)
    assert data2.time.value[0] == 1.
    assert data2.time.value.shape[0] == 2

def test_split_gaps():
    
    # Run integration test that succeeds
    gaps = [(0, 20.), (21., 34), (37, 41)]
    splits = [1.5, 14., 39.]
    result = [(0, 1.5), (1.5, 14.0), (14.0, 20.0),
              (21.0, 34), (37, 39.0), (39.0, 41)]
    
    assert split_gaps(gaps, splits) == result
    
    # Run integration test that succeeds with the default input
    # should return the input 
    gaps = [(0, 20.), (21., 34), (37, 41)]
    splits = []
    
    assert split_gaps(gaps, splits) == gaps
    

    # Get instruction on how to pick good split values
    # when you mess it up:

    gaps = [(0, 20.), (21., 34), (37, 41)]


    # by passing NaN
    with pytest.raises(IndexError):
        splits = [1.5, np.nan, 39.]
        split_gaps(gaps, splits)

    # by passing inf
    with pytest.raises(IndexError):
        splits = [1.5, np.inf, 39.]
        split_gaps(gaps, splits)

    # by passing values outside of range
    with pytest.raises(IndexError):
        splits = [1.5, 99., 39.]
        split_gaps(gaps, splits)



import pytest
import copy

import numpy as np
import pandas as pd

import logging

from ..utils import generate_random_power_law_distribution
from ..wheatland import generate_fake_data
from ..ffd import (FFD,
                   _get_multistar_factors,
                   _ML_powerlaw_estimator,
                   _de_biased_upper_limit,
                   _de_bias_alpha,
                   _calculate_percentile_max_ed,
                   _calculate_max_ed,
                   _stabilised_KS_statistic,
                   _calculate_KS_acceptance_limit,
                   _apply_stabilising_transformation,
                    )

#------------------------------ TESTING FFD() ------------------------------------------

def test_init_FFD():
    
    # Generate a flare table
    a, b, g, size = 10, 1e3, -1, 200
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)
    df = pd.DataFrame({"ed_rec": pwl,
                       "ed_corr": pwl * 1.2,
                       "recovery_probability":.8,
                       "TIC":list(np.arange(1,21))*10
                       })
    
    # init an FFD object
    ffd = FFD(df)
    
    # check correct initialisation
    assert ffd.f.shape[0] == 200
    assert (ffd.f.columns.values == ['ed_rec', 'ed_corr',
                                     'recovery_probability', 
                                     'TIC']).all()

#---------------------------------------------------------------------------------------

#--------------------- TESTING SETTER OUTPUTS IN PROPERTIES ----------------------------

def test_setter_outputs(caplog):
    """Test all setters to work in both silent and loud modes.
    Using the caplog fixture from pytest to get logging outputs.""" 
    with caplog.at_level(logging.INFO):
        simple_ffd = FFD()
    print(caplog)

    assert ('No total observing time given. Set to 1. '
            'You are now working with number counts instead of frequency.' in caplog.text)

    # multiple_stars
    with caplog.at_level(logging.INFO):
        simple_ffd._multiple_stars = True
    assert ('Setting multiple_stars flag with True.' in caplog.text) == False
    with caplog.at_level(logging.INFO):    
        simple_ffd.multiple_stars = True
    assert 'Setting multiple_stars flag with True.' in caplog.text
    
    a = [1,2,3]

    # ed
    s = f"Setting ED with new values, size {len(a)}."
    with caplog.at_level(logging.INFO):
        simple_ffd._ed = a
    assert (s in caplog.text) == False
    with caplog.at_level(logging.INFO):    
        simple_ffd.ed = a
    assert s in caplog.text

    # freq
    s = f"Setting frequency values with new values, size {len(a)}."
    with caplog.at_level(logging.INFO):    
        simple_ffd._freq = a
    assert (s in caplog.text) == False
    with caplog.at_level(logging.INFO):
        simple_ffd.freq = a
    assert s in caplog.text

    # count_ed
    with caplog.at_level(logging.INFO):
        simple_ffd._count_ed = a
    s = (f"Setting frequency adjusted count values "
         f"with new values, size {len(a)}.")
    assert (s in caplog.text) == False
    with caplog.at_level(logging.INFO):    
        simple_ffd.count_ed = a
    assert s in caplog.text
    
#---------------------------------------------------------------------------------------

#--------------------- TESTING ed_and_freq() ---------------------

def test_ed_and_freq():
    """Tests _ed_and_counts under the hood."""
    
    # Generate a flare table
    a, b, g, size = 10, 1e3, -1, 200
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)
    df = pd.DataFrame({"ed_rec":pwl,
                   "ed_corr":pwl*1.2,
                   "recovery_probability":.8,
                   "TIC":list(np.arange(1,21))*10
                    })
    
    # Init an FFD object
    ffd = FFD(df)

    # Check if the results are correct for all combinations of parameters:
    
    
    # SINGLE STAR SAMPLE
    # -----------------------------------------------------------------
    
    #------------------------------------------------------------------
    # no correction
    ed, freq, counts = ffd.ed_and_freq(energy_correction=False,
                               recovery_probability_correction=False,
                                multiple_stars=False)
    assert (freq == np.arange(1,201)).all()
    assert (ed == df.ed_rec.sort_values(ascending=False)).all()
    # Default setting ...
    assert ffd.tot_obs_time == 1.
    # ... implies that smallest frequency = 1
    assert freq[0] == 1
    #------------------------------------------------------------------
    
    #------------------------------------------------------------------
    # only energy correction
    ed, freq, counts = ffd.ed_and_freq(energy_correction=True,
                               recovery_probability_correction=False,
                                multiple_stars=False)
    assert (freq == np.arange(1,201)).all()
    assert (ed == df.ed_corr.sort_values(ascending=False)).all()
    # Default setting ...
    assert ffd.tot_obs_time == 1.
    # ... implies that smallest frequency = 1
    assert freq[0] == 1
    #------------------------------------------------------------------
    
    #------------------------------------------------------------------
    # energy and frequency correction
    ed, freq, counts = ffd.ed_and_freq(energy_correction=True,
                               recovery_probability_correction=True,
                                multiple_stars=False)

    assert (ed == df.ed_corr.sort_values(ascending=False)).all()
    assert (freq == np.arange(1,201)/.8).all()
    # Default setting ...
    assert ffd.tot_obs_time == 1.
    # ... implies that smallest frequency = 1 / recovery probablity
    assert freq[0] == 1 / .8
    #------------------------------------------------------------------

    # MUTLIPLE STARS IN SAMPLE
    # -------------------------------------------------------------------

    ffd = FFD(df)

    # You must pass a Key to ID
    with pytest.raises(KeyError):
        ed, freq, counts = ffd.ed_and_freq(energy_correction=False,
                               recovery_probability_correction=False,
                                multiple_stars=True)

    ffd = FFD(df,ID="TIC")
    
    #------------------------------------------------------------------
    # no correction
    ed, freq, counts = ffd.ed_and_freq(energy_correction=False,
                               recovery_probability_correction=False,
                                multiple_stars=True)
    #frequencies always increase in FFD
    assert (np.diff(freq) > 0.).all()
    #ED array is the same as the ed_rec column
    assert (ed == df.ed_rec.sort_values(ascending=False)).all()
    # When multiple stars are involved with different minimum
    # detected energies, the frequency adjusted ED array must be larger
    # than the original one
    assert len(ffd.count_ed) > len(ffd.freq)
    # The ultiple_stars flag must be set!
    assert ffd.multiple_stars
    # Default setting ...
    assert ffd.tot_obs_time == 1.
    # ... implies that smallest frequency = 1
    assert freq[0] == 1
    
    #------------------------------------------------------------------
    # only energy correction
    ed, freq, counts = ffd.ed_and_freq(energy_correction=True,
                               recovery_probability_correction=False,
                                multiple_stars=True)

    #frequencies always increase in FFD
    assert (np.diff(freq) > 0.).all()
    #ED array is the same as the ed_corr column
    assert (ed == df.ed_corr.sort_values(ascending=False)).all()
    # When multiple stars are involved with different minimum
    # detected energies, the frequency adjusted ED array must be larger
    # than the original one
    assert len(ffd.count_ed) > len(ffd.freq)
    # The ultiple_stars flag must be set!
    assert ffd.multiple_stars
    # Default setting ...
    assert ffd.tot_obs_time == 1.
    # ... implies that smallest frequency = 1
    assert freq[0] == 1
    _f = copy.copy(freq)
    #------------------------------------------------------------------

    #------------------------------------------------------------------
    # energy and frequency correction
    ed, freq, counts = ffd.ed_and_freq(energy_correction=True,
                               recovery_probability_correction=True,
                                multiple_stars=True)

    #frequencies always increase in FFD
    assert (np.diff(freq) > 0.).all()
    #ED array is the same as the ed_rec column
    assert (ed == df.ed_corr.sort_values(ascending=False)).all()
    # When multiple stars are involved with different minimum
    # detected energies, the frequency adjusted ED array must be larger
    # than the original one
    assert len(ffd.count_ed) > len(ffd.freq)
    # The ultiple_stars flag must be set!
    assert ffd.multiple_stars
    # Default setting ...
    assert ffd.tot_obs_time == 1.
    # ... implies that smallest frequency = 1 / recovery_probability
    assert freq[0] == 1 / .8
    # Adjusting for recovery probability gives a values divided
    # by it
    assert _f/.8 == pytest.approx(freq)
    #------------------------------------------------------------------
    
    # Check failing case:
    # -------------------------------------------------------------
    
    # You must set energy_correction if you want to use 
    # recovery_probability correction
    with pytest.raises(KeyError):
        ed, freq, counts = ffd.ed_and_freq(energy_correction=False,
                               recovery_probability_correction=True,
                                multiple_stars=True)
        
    with pytest.raises(KeyError):
        ed, freq, counts = ffd.ed_and_freq(energy_correction=False,
                               recovery_probability_correction=True,
                                multiple_stars=False)

#---------------------------------------------------------------------------------------

#--------------------------- TESTING fit_powerlaw() -------------------------------
        
def test_fit_mmle_powerlaw():
    # Generate a flare table
    a, b, g, size = 10, 1e3, -1, 200
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)
    df = pd.DataFrame({"ed_rec":pwl,
                       "ed_corr":pwl*1.2,
                       "recovery_probability":.8,
                       "TIC":list(np.arange(1,21))*(size//20)
                       })

    # init an FFD object
    ffd = FFD(f=df)
    ed, freq, counts = ffd.ed_and_freq()
    
    # mode 'mmle'
    ffd.fit_powerlaw("mmle")
    assert (1.963553855895996, 0.08012203082491737) == (ffd.alpha, ffd.alpha_err)
    assert (1753.1677389526367, 140.50464845948764) == (ffd.beta, ffd.beta_err)

    # mode 'mcmc'
    ffd.fit_powerlaw("mcmc")
    assert ((2.041785081531872, 0.07598158206969874, 0.07111524587912488) ==
            (ffd.alpha, ffd.alpha_up_err, ffd.alpha_low_err))
    assert ((2302.96755340424, 670.4364587925747, 503.09507205326236) ==
            (ffd.beta, ffd.beta_up_err, ffd.beta_low_err))

#---------------------------------------------------------------------------------------

#--------------------------- TESTING fit_mmle_powerlaw() -------------------------------
        
def test_fit_mmle_powerlaw():
    # Generate a flare table
    a, b, g, size = 10, 1e3, -1, 200
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)
    df = pd.DataFrame({"ed_rec":pwl,
                       "ed_corr":pwl*1.2,
                       "recovery_probability":.8,
                       "TIC":list(np.arange(1,21))*(size//20)
                       })

    # init an FFD object
    ffd = FFD(f=df)
    ed, freq, counts = ffd.ed_and_freq()
    
    # check the result
    assert (1.963553855895996, 0.08012203082491737) == ffd.fit_mmle_powerlaw()

#---------------------------------------------------------------------------------------

#------------------------------ TESTING is_powerlaw_truncated() ------------------------

cases = [(1000, False), (900, False), (800, False),
         (200, True), (20, True)]

@pytest.mark.parametrize("l,i", cases)
def test_is_powerlaw_truncated(l,i):
    a, b, g, size = 10, 1e3, -1, 200
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)
    f = pd.DataFrame({"ed_rec":pwl})
    simple_truncated_ffd = FFD(f=f[f.ed_rec < l])#truncate at l seconds 
    simple_truncated_ffd.alpha = 2.
    ed, freq, counts = simple_truncated_ffd.ed_and_freq()
    assert simple_truncated_ffd.is_powerlaw_truncated() == i
    
#---------------------------------------------------------------------------------------

#------------------------------ TESTING is_powerlaw() ----------------------------------

def test_is_powerlaw():

    a, b, g, size = 10, 1e3, -1., 200
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)

    # init an FFD object
    ffd = FFD(f=pd.DataFrame({"ed_rec":pwl}))
    ffd.ed_and_freq()
    
    ffd.alpha = 2.
    # pwl is a power law with exponent 2
    assert ffd.is_powerlaw()

    ffd.alpha = 2.3
    # pwl is not a power law with exponent 2
    assert not ffd.is_powerlaw()


    a, b, g, size = 10, 1e3, -1., 20
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)

    # init an FFD object
    ffd = FFD(f=pd.DataFrame({"ed_rec":pwl}))
    ffd.ed_and_freq()

    ffd.alpha = 2.
    # pwl is a power law with exponent 2
    assert ffd.is_powerlaw()

    ffd.alpha = 2.3
    # pwl is not a power law with exponent 2 but 20 is too small of a sample
    assert ffd.is_powerlaw()

    a, b, g, size = 10, 1e3, -1., 50
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)
    ffd = FFD(f=pd.DataFrame({"ed_rec":pwl}))

    with pytest.raises(TypeError): #throw error when alpha is missing
        ffd.is_powerlaw()

#---------------------------------------------------------------------------------------

#------------------------------ TESTING _get_ed() ----------------------------------

def test__get_ed():

    a, b, g, size = 10, 1e3, -1., 200
    pwl = generate_random_power_law_distribution(a, b, g, size=size, seed=80)

    # init an FFD object

    # case multiple_stars==False
    ffd = FFD(f=pd.DataFrame({"ed_rec":pwl}))
    ffd.ed_and_freq()
    
    assert (ffd._get_ed() == np.sort(pwl)[::-1]).all()

    # case multiple_stars==True
    pwl = np.sort(pwl)[::-1]
    ffd = FFD(f=pd.DataFrame({"ed_rec":pwl, "ID": [1]*(len(pwl)-1) + [2]}),
              ID="ID")
    ffd.ed_and_freq(multiple_stars=True)
    assert (ffd._get_ed() == list(pwl) + [np.min(pwl)]).all()


# -------------------- TESTING fit_mcmc_powerlaw() -------------------------------------

def test_fit_mcmc_powerlaw():

    # ----------------------------------------------------------
    # Run an integration test:

    # Generate fake FFD dictionary
    fake = generate_fake_data(13, 120, 5, 1.7, seed=9000)

    # Create an instance of an FFD
    ffd = FFD()
    ffd.f = pd.DataFrame({"ed_rec":fake["events"]})
    ffd.tot_obs_time = fake['Tprime'] 
    ffd.beta_prior = fake["beta_prior"]
    ffd.alpha_prior = fake["alpha_prior"]
    ed, freq, counts = ffd.ed_and_freq()

    # Call the function you wish to test
    np.random.seed(42)
    BFA = ffd.fit_mcmc_powerlaw()

    # Keep this for de-bugging
    fig = BFA.show_corner_plot()

    # check if the results are approximately right
    assert ffd.alpha == pytest.approx(1.7, rel=.1) 
    assert ffd.beta == pytest.approx(5., rel=.2)
    assert ffd.beta_err is None
    assert ffd.alpha_err == pytest.approx(0.075, rel=.1)
    assert ffd.alpha_up_err == pytest.approx(0.080, rel=.1)
    assert ffd.alpha_low_err == pytest.approx(0.073, rel=.1)
    assert ffd.beta_up_err == pytest.approx(.827, rel=.1)
    assert ffd.beta_low_err == pytest.approx(.720, rel=.1)
    assert ffd.beta_prior == 5.
    assert ffd.alpha_prior == 1.7
    assert len(ffd.ed) == len(fake["events"])
    
    # ----------------------------------------------------------
    # Run a failing test

    # This should because ffd.f is not given:
    ffd = FFD()
    with pytest.raises(ValueError):
        BFA = ffd.fit_mcmc_powerlaw()

#---------------------------------------------------------------------------------------

# -------------------- TESTING _get_multistar_factors() ---------------------------------

def test__get_multistar_factors():
    
    # Generate a flare table
    N = 20
    testdf = pd.DataFrame({"ID":np.arange(N)%4,
                           "sortcol":np.arange(200, 200-N, -1)})
    
    # call _get_multistar_factors
    f = _get_multistar_factors(testdf, "ID", "sortcol")
    
    # Check if the result is as expected
    assert (f == np.array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
                           1., 1., 1., 1., 1., 1., 0.75, 0.5, 0.25])).all()

    # If the required keys don't exist throw errors
    with pytest.raises(KeyError):
        testdf = pd.DataFrame({"ID":np.arange(N)%4})
        f = _get_multistar_factors(testdf, "ID", "sortcol")

    with pytest.raises(KeyError):
        testdf = pd.DataFrame({"sortcol":np.arange(200, 200-N, -1)})
        f = _get_multistar_factors(testdf, "ID", "sortcol")

# --------------------------------------------------------------------------------------

#-------------------------- TESTING _ML_powerlaw_estimator() ---------------------------
    
def test__ML_powerlaw_estimator():
    dataformat = [np.array, pd.Series]
    for df in dataformat:
        ed = df([1,1,1,1,2,2,4])
        x0 = .9
        assert np.isnan(_ML_powerlaw_estimator(x0, ed))
        x0 = 1.5
        assert _ML_powerlaw_estimator(x0, ed) == pytest.approx(1.6190804181576444)
        ed = df([])
        with pytest.raises(ValueError):
            _ML_powerlaw_estimator(x0, ed)
        ed = df([-3,2,2,2])
        with pytest.raises(ValueError):
            _ML_powerlaw_estimator(x0, ed)
        ed = df([1,1,1,1])
        with pytest.raises(ValueError):
            _ML_powerlaw_estimator(x0, ed)

#---------------------------------------------------------------------------------------

#------------------------ TESTING _de_biased_upper_limit() -----------------------------

            
def test__de_biased_upper_limit():
    dataformat = [np.array, pd.Series]
    for df in dataformat:
        data = df([])
        with pytest.raises(ValueError):
            _de_biased_upper_limit(data, 2.)
        data = df([1,10, 100])
        assert _de_biased_upper_limit(data, 1000000.) == pytest.approx(100., rel=1e-4)
        data = df([1,1,1])
        with pytest.raises(ValueError):
            _de_biased_upper_limit(data, 3.)
        data = df([-1,1,1])
        with pytest.raises(ValueError):
            _de_biased_upper_limit(data, 3.)

#---------------------------------------------------------------------------------------

#----------------------------- TESTING _de_bias_alpha() --------------------------------
            
def test__de_bias_alpha():
    assert _de_bias_alpha(200,1) == 1.
    with pytest.raises(ValueError):
        _de_bias_alpha(np.nan,2)
    with pytest.raises(ValueError):
        _de_bias_alpha(30,np.nan)
    with pytest.raises(ValueError):
        _de_bias_alpha(np.nan,np.nan)
    with pytest.raises(ZeroDivisionError):
        _de_bias_alpha(2,2)
        
#---------------------------------------------------------------------------------------

#-------------------------- TESTING _calculate_max_ed() --------------------------------

def test__calculate_max_ed():
    # fake data
    data = np.linspace(10,1e3,100)
    alpha = 2.

    # apply function once
    maxval = _calculate_max_ed(data, alpha, maxlim=1e8, seed=10)

    assert maxval == pytest.approx(808.1118136567375)

    # apply it again

    maxval = _calculate_max_ed(data, alpha, maxlim=1e8, seed=2000)
    assert maxval == pytest.approx(1372.7172156831662)

    # test failure mode alpha or data is not finite
    with pytest.raises(TypeError):
        maxval = _calculate_max_ed(data, None, maxlim=1e8, seed=2000)

    with pytest.raises(TypeError):
        maxval = _calculate_max_ed(None, alpha, maxlim=1e8, seed=2000)

    with pytest.raises(AssertionError):
        maxval = _calculate_max_ed(data, np.nan, maxlim=1e8, seed=2000)

    with pytest.raises(AssertionError):
        maxval = _calculate_max_ed([np.nan, 1., 2., 3. ,4.],
                                   alpha, maxlim=1e8, seed=2000)

#---------------------------------------------------------------------------------------

#---------------------- TESTING _calculate_percentile_max_ed() -------------------------

def test__calculate_percentile_max_ed():
    # fake data
    data = np.linspace(10,1e3,100)
    alpha = 2.
    n = 10000
    percentile = 3.

    # apply function
    vals, edcrit = _calculate_percentile_max_ed(data, alpha, n, percentile)
    assert len(vals) == n
    assert edcrit == pytest.approx(289, abs=20)
    
    # test failure modes
    with pytest.raises(AssertionError):
        vals, edcrit = _calculate_percentile_max_ed(data, alpha, n, 200)
    with pytest.raises(AssertionError):
        vals, edcrit = _calculate_percentile_max_ed(data, alpha, n, -3)
    with pytest.raises(AssertionError):
        vals, edcrit = _calculate_percentile_max_ed(data, alpha, n, np.nan)
        
#---------------------------------------------------------------------------------------

#------------------------ TESTING _stabilised_KS_statistic() -----------------------------

def test__stabilised_KS_statistic():
    sizes = [1e2,1e3,1e4]
    minval, maxval = 10, 1e4
    datas = [generate_random_power_law_distribution(minval, maxval, -1., size=int(size),
                                                    seed=10) for size in sizes]
    
    KSlist = [_stabilised_KS_statistic(data, 2., False) for data in datas]
    assert KSlist[0] > KSlist[1]
    assert KSlist[1] > KSlist[2]
    

#---------------------------------------------------------------------------------------

#---------------------- TESTING _calculate_KS_acceptance_limit() -----------------------

def test__calculate_KS_acceptance_limit():
    with pytest.raises(ValueError):
        _calculate_KS_acceptance_limit(0, sig_level=0.05)
    with pytest.raises(ValueError):
        _calculate_KS_acceptance_limit(0, sig_level=-0.05)
    with pytest.raises(ValueError):
        _calculate_KS_acceptance_limit(0, sig_level=1.05)
    assert (_calculate_KS_acceptance_limit(100, sig_level=0.05)
            == pytest.approx(0.13581, rel=1e-4))
    assert (_calculate_KS_acceptance_limit(100, sig_level=0.01)
            == pytest.approx(0.16276, rel=1e-4))
    assert (_calculate_KS_acceptance_limit(100, sig_level=0.01)
            > _calculate_KS_acceptance_limit(1000, sig_level=0.01))
    
#---------------------------------------------------------------------------------------

#--------------------- TESTING _apply_stabilising_transformation() ---------------------

def test__apply_stabilising_transformation():
    u = [.1,.2,.3]
    assert _apply_stabilising_transformation(u).shape[0] == 3
    assert (np.diff(u) > 0).all()

    # giving an empty arrays returns an empty array
    assert (_apply_stabilising_transformation([]) == np.array([])).all()

    # Passing negative values throws an error
    with pytest.raises(ValueError):
        _apply_stabilising_transformation([-1.,.2,.4])



import numpy as np
import pytest

from ..altai import (find_flares,
                     find_flares_in_cont_obs_period,
                     chi_square,
                     equivalent_duration,
                     find_iterative_median)

from ..flarelc import FlareLightCurve
from .test_flarelc import mock_flc

def test_detrend_savgol():
    """Test if different window_length kwargs are processed correctly."""
    
    N = int(1e4)
    time = np.linspace(2000,2050,N)
    flux = np.sin(time / 2.) * 30. + 5e4 + np.random.rand(N) * 35. + 5e-4 * ((time-2004.)**3 - 30 * (time-2004)**2)
    flux[5000:5010] = flux[5000:5010] + np.array([500,250,150,80,60,30,20,10,7,4])
    flux[4500:4809] = np.nan
    flux_err = np.random.rand(N) * 10.
    flc = FlareLightCurve(targetid=10000009, time=time, flux=flux, flux_err=flux_err)
    
    flcds = [flc.detrend("savgol"),
             flc.detrend("savgol", window_length=201),
             flc.detrend("savgol", window_length=(101,205)),
             flc.detrend("savgol", window_length=[101,205])]
    for flcd in flcds:
            assert flcd.detrended_flux.value.shape[0] == 1e4-309
            
    N = int(1e4)
    time = np.linspace(2000,2050,N)
    np.random.seed(200)
    flux = np.sin(time / .03) * 30. + 5e4 + np.random.rand(N) * 25. + 5e-4 * ((time-2004.)**3 - 30 * (time-2004)**2)
    flux[5000:5010] = flux[5000:5010] + np.array([500,250,150,80,60,30,20,10,7,4])
    flux[4500:4809] = np.nan
    flux_err = np.random.rand(N) * 35. # this reflects the real noise
    flc = FlareLightCurve(targetid=10000009, time=time, flux=flux, flux_err=flux_err)
    
    flcds = [flc.detrend("savgol"),
             flc.detrend("savgol", window_length=201),
             flc.detrend("savgol", window_length=(101,205)),
             flc.detrend("savgol", window_length=[25,25])]
    for flcd in flcds:
            assert flcd.detrended_flux.value.shape[0] == 1e4-309
            
    # The last de-trending iteration is the only appropriate one to give good 
    # results given the rapid variability of the light curve. So let's check
    # the outcome of this one. It should only recover the one flare we intro-
    # duced above around t=2025
    flares = flcd.find_flares().flares
    
    f = flares.iloc[0,:]
    assert f.tstart == pytest.approx(2025, abs=5e-3)
    assert f.ed_rec == pytest.approx(8.31358,abs=f.ed_rec_err)
    assert f.istart == 4691
    assert f.istop == 4695
    assert f.total_n_valid_data_points == 1e4-309
    assert f.dur == pytest.approx(f.tstop - f.tstart, rel=1e-4)

    
def test_iterative_median():

    flc = mock_flc(detrended=True)
    lc1 = find_iterative_median(flc, n=1)
    lc2 = find_iterative_median(flc, n=2)

    # test that find_iterative_median converges after one iteration for mock FLC
    assert np.median(flc.it_med.value) != np.median(lc1.it_med.value)
    assert np.median(lc1.it_med.value) == np.median(lc2.it_med.value)

def test_find_flares():
    """
    Integration test of a mock example light curve is given in test_flarelc.
    Add unit tests!
    """
    flc = mock_flc(detrended=False)
    with pytest.raises(TypeError):
        #raises error bc find_flares only works on detrended_flux
        find_flares(flc)
     
    # Check if all columns are created
    flc = mock_flc(detrended=True)
    for col in  ['istart', 'istop', 'cstart', 'cstop', 'tstart',
                 'tstop', 'ed_rec', 'ed_rec_err', 'ampl_rec', 
                 'total_n_valid_data_points', 'dur']:
        assert col in flc.flares.columns

def test_find_flares_in_cont_obs_period():
    """
    Integration test of a mock example light curve is given in test_flarelc.
    Add unit tests!
    """
    # Set up a light curve with a flare
    
    # time series (not needed but helpful for debugging)
    time = np.linspace(30,40,100)
    
    # flux array
    flux = 400 + np.random.normal(0,30,100)
    
    # add the flare
    flux[30:41] = [800, 660, 500, 490, 480, 470, 460, 455, 450, 445, 440]
    
    # give the LC characteristics
    median, error, sigma = 400., 1., 30.,

    # find the flare
    isflare = find_flares_in_cont_obs_period(flux, median, error, 
                                             sigma=sigma, N1=3, N2=2,
                                             N3=3, addtail=False)
    # check if found
    assert (np.where(isflare)[0] == np.array([30, 31, 32, 33])).all()

    # add flare decay phase to search
    isflare = find_flares_in_cont_obs_period(flux, median, error, 
                                             sigma=sigma, N1=3, N2=2,
                                             N3=3, addtail=True)


    # check if found
    assert (np.where(isflare)[0] == np.array([30, 31, 32, 33, 34, 35, 36])).all()

    # add a second flare
    flux[40:49] = [510, 500, 491, 470, 460, 455, 450, 445, 440]

    # find with tails
    isflare = find_flares_in_cont_obs_period(flux, median, error, 
                                             sigma=sigma, N1=3, N2=2, 
                                             N3=3, addtail=True)

    # check if found
    assert (np.where(isflare)[0] ==  np.array([30, 31, 32, 33, 34, 35, 
											   36, 40, 41, 42, 43, 
											   44])).all()

    # bad value for tailthreshdiff throws error
    with pytest.raises(ValueError):
        isflare = find_flares_in_cont_obs_period(flux, median, error, 
                                             sigma=sigma, N1=3, N2=2, 
                                             N3=3, addtail=True, tailthreshdiff=3)




def test_chi_square():
    """Test an abvious example"""
    residual = np.full(5,1.)
    error = np.full(5,.1)
    assert chi_square(residual, error) == 100.

def test_equivalent_duration():
    """Test a triangle shaped flare in a toy light curve."""
    detrended_flux = np.full(1000,1.)
    detrended_flux[60:70] = np.array([10,9,8,7,6,5,4,3,2,1])
    lc = FlareLightCurve(time=np.arange(1000)/86400.)
    lc.detrended_flux=detrended_flux
    lc.it_med =np.full(1000,1.)
    lc.detrended_flux_err=np.full(1000,1e-8)
   # print(lc.saturation)
    ed, ed_err = equivalent_duration(lc, 60, 70, err=True)
    assert ed == pytest.approx(45,rel=1e-8)
    assert ed_err == pytest.approx(2.665569e-08, rel=1e-4)



import os
import pytest

import numpy as np
import pandas as pd

import matplotlib

from ..injrecanalysis import (characterize_flares,
                              tile_up_injection_recovery,
                              multiindex_into_df_with_nans,
                              percentile,
                              wrap_characterization_of_flares,
                              plot_heatmap,
                              _heatmap,
                              setup_bins,
                             )
from ..flarelc import FlareLightCurve
                             
from . import PACKAGEDIR                             

def test_wrap_characterization_of_flares():
    # Create a fake injection-recovery data set

    np.random.seed(20) #seed the random number generator
    n=int(1e5) # size

    amplrec = np.random.rand(n)*5
    amplitude = amplrec*1.2

    durrec = np.random.rand(n)*.2
    durrecsample = (durrec // (2. / 60. / 24.)) * (2. / 60. / 24.)
    tstart = np.random.rand(n) * 12000
    tstop = tstart + durrecsample
    dur = durrec*.5

    edinj = np.random.rand(n)*5000
    edrec = edinj*.8
    edrec[np.where(np.random.rand(n) < 0.1)] = 0.

    injrec = pd.DataFrame({"ampl_rec":amplrec,
                            "amplitude":amplitude,
                            "duration_d":dur,
                            "ed_inj":edinj,
                            "ed_rec":edrec,
                            "ed_rec_err":edrec*.1,
                            "dur":durrecsample,
                            "tstart":tstart,
                            "tstop":tstop,
                            })

    # read in test flares with realistic values from file:
    testflares = pd.read_csv(os.path.join(PACKAGEDIR,
                                          "tests/testfiles/testflares.csv"))

    #---------------------------------------------------------
    # Call the function
    fl = wrap_characterization_of_flares(injrec, testflares, ampl_bins=70, dur_bins=160)


    #---------------------------------------------------------
    # Check some numbers

    assert fl.shape[0] == 5
    assert (fl.ed_rec.values / 0.8 == fl.ed_corr.values).all()
    assert (fl.ed_ratio_count.values < 15).all()  # the numbers in the bins are not too high
    assert (fl.ed_ratio_count.values > 5).all()  # the numbers in the bins are not too low
    assert (fl.recovery_probability_count.values < 30).all()  # the numbers in the bins are not too high
    assert (fl.recovery_probability_count.values > 15).all()  # the numbers in the bins are not too low
    assert len(np.where(fl.isnull())[0]) == 0 # no NaNs
    assert (fl.ed_ratio.values == 0.8).all() # should be by design
    assert (fl.amplitude_ratio * 1.2 ==  1.).all() # should be by design
    # recovery probability is .9 for injected but there are shifts induced to this values by ED ratio correction:
    assert (fl.recovery_probability.values > .75).all()
    assert (fl.ed_corr_err.values ==
            pytest.approx(np.array([6.04627777e+02, 1.06333934e+02,
                                    5.90000000e-01, 4.30000000e-01,
                                    5.80000000e-01])))
    assert fl.amplitude_corr_err.values == pytest.approx(0) # shoulb be by data set design
    assert (fl.duration_corr_err.values < (2. / 60. / 24.)).all() #uncertainties smaller than time resolution by design


def test_characterize_flares():
    #-------------------------------
    # Create a fake data set
    # First injection recovery data:

    np.random.seed(20)
    n=int(1e5)
    amplrec = np.random.rand(n)*10
    amplitude = amplrec*1.2
    durrec = np.random.rand(n)*.6
    dur = durrec*.5
    edinj = np.random.rand(n)*5000
    edrec = edinj*.8
    dftest0 = pd.DataFrame({"ampl_rec":amplrec,
                            "amplitude":amplitude,
                            "duration_d":dur,
                            "ed_inj":edinj,
                            "ed_rec":edrec,
                            "dur":durrec,
                            "rec":(np.random.rand(n) > 0.1).astype(float)})

    # Then, flare detections:

    n=int(10)
    amplrec = np.random.rand(n)*10
    edrec = np.random.rand(n)*4000
    tstop = np.random.rand(n)*2000.
    tstart = tstop - .3
    testflares = pd.DataFrame({"ampl_rec":amplrec,
                               "ed_rec":edrec,
                               "tstop": tstop,
                               "tstart": tstart})
    #-------------------------------
    # Call the function:

    fl, ds = characterize_flares(testflares, dftest0,
                             ampl_bins=np.linspace(0, 12, 11),
                             dur_bins=np.linspace(0, .6, 11))

    #-------------------------------
    # Do some tests:

    assert fl.recovery_probability.values == pytest.approx(.9, rel=.05)
    assert (fl.duration_ratio.values == 2.).all()
    assert (fl.amplitude_ratio.values == 10. / 12.).all()
    assert (fl.ed_ratio.values == .8).all()
    assert (fl.amplitude_corr * fl.amplitude_ratio == fl.ampl_rec).values.all()
    assert (fl.duration_corr * fl.duration_ratio == fl.dur).values.all()
    assert (fl.ed_corr * fl.ed_ratio).values == pytest.approx(fl.ed_rec, rel=.05)
    assert fl.shape[1] == 16
    assert fl.shape[0] == 10
    assert len(ds.keys()) == 4
    assert fl.recovery_probability_count.values == pytest.approx(2000, rel=.05)

def test_tile_up_injection_recovery():
    np.random.seed(20)
    n=int(1e5)
    amplrec = np.random.rand(n)*10
    amplitude = amplrec*1.2
    durrec = np.random.rand(n)*.6
    dur = durrec*.5
    edinj = np.random.rand(n)*5000
    edrec = edinj*.8
    dftest0 = pd.DataFrame({"ampl_rec":amplrec,
                           "amplitude":amplitude,
                          "duration_d":dur,
                          "ed_inj":edinj,
                          "ed_rec":edrec,
                          "dur":durrec,
                          "rec":(np.random.rand(n) > 0.5).astype(float)})


    for t, vali, res in [("recovery_probability", "rec", .5),
                   ("ed_ratio", "edrat", .8),
                   ("amplitude_ratio", "amplrat", 10./12.),
                   ("duration_ratio","durrat", 2)]:
        resdf, val = tile_up_injection_recovery(dftest0, t, ampl="amplitude", dur="duration_d",
                                       otherfunc = "count",
                                       ampl_bins=np.linspace(0, 12, 11),
                                       dur_bins=np.linspace(0, .3, 11))

        assert resdf["count"].values.shape[0] == 100
        assert resdf["count"].mean() == 1000.
        assert val == vali
        assert resdf[val].median() == pytest.approx(res, rel=1e-2)

    with pytest.raises((KeyError, AttributeError)):
        dftest1 = pd.DataFrame()
        resdf, val = tile_up_injection_recovery(dftest1, "recovery_probability")

    dftest0.loc[dftest0.rec==0.,["ampl_rec","dur","ed_rec"]] = np.nan
    resdf, val = tile_up_injection_recovery(dftest0, "recovery_probability")

    for t, vali, res, counts in [("recovery_probability", "rec", .5, 1000),
                   ("ed_ratio", "edrat", .8, 500),
                   ("amplitude_ratio", "amplrat", 10./12., 500),
                   ("duration_ratio","durrat", 2, 500)]:
        resdf, val = tile_up_injection_recovery(dftest0, t, ampl="amplitude", dur="duration_d",
                                       otherfunc = "count",
                                       ampl_bins=np.linspace(0, 12, 11),
                                       dur_bins=np.linspace(0, .3, 11))
        print(t, resdf["count"].median())
        assert resdf["count"].values.shape[0] == 100
        assert resdf["count"].median() == pytest.approx(counts, rel=1e-2)
        assert val == vali
        assert resdf[val].median() == pytest.approx(res, rel=1e-2)


def test_multiindex_into_df_with_nans():
    np.random.seed(30)
    i1 = pd.interval_range(start=0, end=5)
    i2 = pd.interval_range(start=0, end=8)
    index = pd.MultiIndex.from_product([i1,i2], names=['ampl_rec', 'dur'])
    testdf = pd.DataFrame(np.random.rand(40), index=index, columns=["edrat"])
    s = pd.Series({"ampl_rec":.1, "dur":0.5})
    assert multiindex_into_df_with_nans(s, testdf) == pytest.approx(.6441435606)
    for v1,v2 in [(np.nan, .5), (1., np.nan),
                  (np.nan, np.nan), (6,3), (9,9)]:
        s = pd.Series({"ampl_rec":v1, "dur":v2})
        assert np.isnan(multiindex_into_df_with_nans(s, testdf))


def test_percentile():
    s = pd.Series(np.linspace(0,10,11), name="smth")
    assert percentile(s,50) == 5.
    assert percentile(s,16) == 1.6
    assert np.isnan(percentile(s, np.nan))
    s = pd.Series(np.zeros(10), name="smth")
    assert percentile(s,50) == percentile(s,70)
    s[:] = np.nan
    assert np.isnan(percentile(s,16))
    s[2] = 1.
    assert percentile(s,1) == percentile(s,99)
    


def test_plot_heatmap():

    df = pd.read_csv(os.path.join(PACKAGEDIR,
                                          "tests/testfiles/injrec_TIC1539914_s9.csv"))
    dft, val = tile_up_injection_recovery(df,"ed_ratio")
    assert isinstance(plot_heatmap(dft, val), matplotlib.figure.Figure)

    dft, val = tile_up_injection_recovery(df,"recovery_probability")
    assert isinstance(plot_heatmap(dft, val), matplotlib.figure.Figure)

    dft, val = tile_up_injection_recovery(df,"amplitude_ratio")
    assert isinstance(plot_heatmap(dft, val), matplotlib.figure.Figure)

    dft, val = tile_up_injection_recovery(df,"duration_ratio")
    assert isinstance(plot_heatmap(dft, val), matplotlib.figure.Figure)



def test__heatmap():
    # Create a minimal empty light curve with an ID
    flcd = FlareLightCurve(targetid="GJ 1243", time=np.linspace(10,1))
    ampl_bins, dur_bins, flares_per_bin = None, None, 20
    
    # If inj-rec data are missing, throw an error.
    with pytest.raises(AttributeError) as err:
        _heatmap(flcd, "recovery_probability", ampl_bins, dur_bins, flares_per_bin)
    
    # Path to test file
    path = "altaipony/tests/testfiles/gj1243_injrec.csv"
    flcd.load_injrec_data(path)
    
    # Test the default case
  #  _heatmap(flcd, "recovery_probability", ampl_bins, dur_bins, flares_per_bin)
    _heatmap(flcd, "ed_ratio", ampl_bins, dur_bins, flares_per_bin)
    
    # Check other cases of bin specifications
    for typ in ["recovery_probability", "ed_ratio"]:
        # only dur_bins
        ampl_bins, dur_bins, flares_per_bin = None, 5, 20
        _heatmap(flcd, typ, ampl_bins, dur_bins, flares_per_bin)

        # only ampl_bins
        ampl_bins, dur_bins, flares_per_bin = 3, None, 20
        _heatmap(flcd, typ, ampl_bins, dur_bins, flares_per_bin)

        # both dur_bins and ampl_bins
        ampl_bins, dur_bins, flares_per_bin = 3, 9, None
        _heatmap(flcd, typ, ampl_bins, dur_bins, flares_per_bin)
        
        # one of the two is defined as array
        ampl_bins, dur_bins, flares_per_bin = 3, np.linspace(5e-4,.004,20), None
        _heatmap(flcd, typ, ampl_bins, dur_bins, flares_per_bin)
        
        # It should also work if you give an array for one bin only
        ampl_bins, dur_bins, flares_per_bin = None, np.linspace(5e-4,.004,20), 7
        _heatmap(flcd, typ, ampl_bins, dur_bins, flares_per_bin)
        
        
def test_setup_bins():
    # Test a number of cases that could possibly be passed.
    
    # A failing example
    # --------------------------------------------------------------
    with pytest.raises(ValueError) as e:
        val = np.linspace(0,130,30)
        injrec = pd.DataFrame({"duration_d": val, "dur": val,
                               "amplitude": val, "ampl_rec": val })
        flares = pd.DataFrame({"dur": val, "ampl_rec": val })
        setup_bins(injrec, flares, ampl_bins=None, 
                   dur_bins=None, flares_per_bin=None)

    # A working example
    # --------------------------------------------------------------

    # Hybrid use case

    # Setup values
    val = np.linspace(0,130,30)
    injrec = pd.DataFrame({"duration_d": val, "dur": val, 
                           "amplitude": val, "ampl_rec": val })
    flares = pd.DataFrame({"dur": val, "ampl_rec": val })


    # Run setup_bins
    a, d = setup_bins(injrec, flares, ampl_bins=np.linspace(0,30,10),
                      dur_bins=3, flares_per_bin=None)


    # Do some checks
    assert len(a) == 10
    assert len(d) == 3
    assert (a == np.linspace(0,30,10)).all()
    assert (d == np.linspace(0,130,3)).all()

    # Another working example
    # --------------------------------------------------------------
    # Hybrid use case

    # Setup values
    val = np.linspace(0,130,300)
    injrec = pd.DataFrame({"duration_d": val, "dur": val, 
                           "amplitude": val, "ampl_rec": val })
    flares = pd.DataFrame({"dur": val, "ampl_rec": val })

    # Run setup_bins
    a, d = setup_bins(injrec, flares, ampl_bins=np.linspace(0,30,10),
                      dur_bins=None, flares_per_bin=3)

    assert len(a) == 10
    assert len(d) == 10
    assert (a == np.linspace(0, 30, len(a))).all()
    assert (d == np.linspace(0, 130, len(d))).all()

    # The lazy example
    # --------------------------------------------------------------
    # Just pass a number of flares per bin

    # Setup values
    val = np.linspace(0,130,300)
    injrec = pd.DataFrame({"duration_d": val, "dur": val, 
                           "amplitude": val, "ampl_rec": val })
    flares = pd.DataFrame({"dur": val, "ampl_rec": val })

    # Run setup_bins
    a, d = setup_bins(injrec, flares, ampl_bins=None,
                      dur_bins=None, flares_per_bin=3)

    # Do some checks
    assert len(a) == 10
    assert len(d) == 10
    assert (a == np.linspace(0, 130, len(a))).all()
    assert (d == np.linspace(0, 130, len(d))).all()

    # The control freak example
    # --------------------------------------------------------------
    # Set the bin edges manually.

    # Setup values
    val = np.linspace(0,130,300)
    injrec = pd.DataFrame({"duration_d": val, "dur": val, 
                           "amplitude": val, "ampl_rec": val })
    flares = pd.DataFrame({"dur": val, "ampl_rec": val })

    # Run setup_bins
    abins, dbins = [2,30,88,210], [3,40,220,780]
    a, d = setup_bins(injrec, flares, ampl_bins=abins,
                      dur_bins=dbins, flares_per_bin=3)

    # Do some checks
    assert len(a) == 4
    assert len(d) == 4

    assert a == abins
    assert d == dbins



import numpy as np
import pandas as pd
import pytest

import os
import warnings


from ..flarelc import FlareLightCurve
from ..altai import find_iterative_median

from .. import PACKAGEDIR
from . import test_ids, test_paths, pathkepler, pathAltaiPony


def mock_flc(origin='TPF', detrended=False, ampl=1., dur=1):
    """
    Mocks a FlareLightCurve with a sinusoid variation and a single positive outlier.

    Parameter
    -----------
    origin : 'TPF' or str
        Mocks a specific origin, such as 'KLC', 'FLC' etc.
    detrended : False or bool
        If False, a sinusoid signal is added to the mock light curve.

    Return
    -------
    FlareLightCurve
    """
    n = 1000
    time = np.arange(0, n/48, 1./48.)
    pixel_time = np.outer(time,np.full((3,3), 1)).reshape((1000,3,3))
    np.random.seed(13854)

    pipeline_mask = np.array([[False, False, False],
                              [False, True,  False],
                              [False, False, False],])
    quality = np.zeros_like(time)
    np.random.seed(33)
    flux_err = np.random.rand(n)/100.
    if detrended==False:
        flux = np.sin(time/2)*7. + 500. +flux_err
        pixel_flux = np.random.rand(len(time),3,3)/100.+500.+np.sin(pixel_time/2)*7.
        pixel_flux_err = np.random.rand(len(time),3,3)/100.
    else:
        flux = 500. + flux_err
        pixel_flux = np.random.rand(len(time),3,3)/100.+500.
        pixel_flux_err = np.random.rand(len(time),3,3)/100.
    flux[15:15+dur] += 500.*ampl
    flux[15+dur:15+2*dur] += 250.*ampl
    flux[15+2*dur:15+3*dur] += 130.*ampl
    flux[15+3*dur:15+4*dur] += 80.*ampl
    quality[17] = 1024
    quality[18] = 128
    keys = {'flux' : flux, 'flux_err' : flux_err, 'time' : time,
            'pos_corr1' : np.zeros(n), 'pos_corr2' : np.zeros(n),
            'cadenceno' : np.arange(n), 'it_med' : np.full_like(time,500.005),
            'quality' : quality,}
    meta = {'targetid' : 800000000,
            'origin' : origin,  'pipeline_mask' : pipeline_mask,
            'pixel_flux' : pixel_flux, 'campaign' : 5, 'ra' : 22.,
            'dec' : 22., 'mission' : 'K2', 'channel' : 55, 
            'pixel_flux_err' : pixel_flux_err, 'time_format': 'bkjd',
            'primary_header':3, 'data_header':2}

    if detrended == False:
        flc = FlareLightCurve(keys, meta=meta)
    else:
        keys["detrended_flux"]=flux
        keys["detrended_flux_err"]=flux_err
        flc = FlareLightCurve(data=keys, meta=meta)
    return flc


def test_get_saturation():
    flc = mock_flc(detrended=True)
    flc.pixel_flux[30,:] = 1e6 # add a saturated point
    
    # Find saturation overall
    r1 = flc.get_saturation()
    assert r1.saturation[30] == True
    assert r1.saturation.shape[0] == flc.flux.value.shape[0]
    assert (r1.saturation[:30] == False).all()
    assert (r1.saturation[31:] == False).all()
    
    
    # Find saturation with level
    r1 = flc.get_saturation(return_level=True)
    assert r1.saturation[30] == 1e6 / 10093
    assert r1.saturation.shape[0] == flc.flux.value.shape[0]
    assert (r1.saturation[:30] == pytest.approx(500 / 10093, rel=1e-4)) # the LC has some noise added per default, so only approximate results
    assert (r1.saturation[31:] == pytest.approx(500 / 10093, rel=1e-4)) # the LC has some noise added per default, so only approximate results
    
    # Find saturation with flares
    flc = flc.find_flares()
    
    #... without saturation attribute present
    r1 = flc.get_saturation()
    assert r1.flares.saturation_f10.iloc[0] == False
    r2 = flc.get_saturation(return_level=True)
    assert r2.flares.saturation_f10.iloc[0] == pytest.approx(0.0495,1e-2)
    r3 = flc.get_saturation(factor=1e-2)
    assert r3.flares['saturation_f0.01'].iloc[0] == True
    
    #... with saturation attribute present
    
    
    # .. that is given as a boolean array
    flc.saturation = np.full(flc.flux.value.shape[0], True)
    r4 = flc.get_saturation()
    assert r4.flares.saturation_f10.iloc[0] == True
    r2 = flc.get_saturation(return_level=True)
    assert r2.flares.saturation_f10.iloc[0] == True # throws a warning, too, test that later
    r3 = flc.get_saturation(factor=1e-2)
    assert r3.flares['saturation_f0.01'].iloc[0] == True # throws a warning, too, test that later
    
    # .. that is given as an array of floats
    flc.saturation = np.full(flc.flux.value.shape[0], 5.0)
    r4 = flc.get_saturation()
    assert r4.flares.saturation_f10.iloc[0] == False
    r2 = flc.get_saturation(return_level=True)
    assert r2.flares.saturation_f10.iloc[0] == 5 
    r3 = flc.get_saturation(factor=1e-2)
    assert r3.flares['saturation_f0.01'].iloc[0] == True 


def test_mark_flagged_flares():
    flc = mock_flc(detrended=True)
    flc = flc.find_flares()
    flc = flc.mark_flagged_flares(explain=True)
    assert flc.flares.quality.iloc[0] == 1152
    s1 = "Sudden sensitivity dropout, Cosmic ray in optimal aperture"
    s2 = "Cosmic ray in optimal aperture, Sudden sensitivity dropout"
    qs = flc.flares.explanation.iloc[0]
    assert ((qs == s1) | (qs == s2))

def test_sample_flare_recovery():
    
    # Generic case
    flc = mock_flc(detrended=True)
    
    flc, fflc = flc.sample_flare_recovery(iterations=2, **{"model":"davenport2014"},)
    #make sure no flares are injected overlapping true flares
    data = flc.fake_flares
    assert data[(data.istart > 14) & (data.istart < 19)].shape[0] == 0
    #test if all injected event are covered in the merged flares:
    assert data.shape[0] == 2
    assert fflc.gaps == [(0, 1000)]
    assert np.median(fflc.it_med.value) == pytest.approx(500.005274113832)
    
    # Custom case
    
    def func(flc):
        flc.detrended_flux =  flc.flux/2.
        flc.detrended_flux_err =  flc.flux_err/2.
        return flc
    
    flc = mock_flc(detrended=True)
    
    flcd, fflc = flc.sample_flare_recovery(iterations=10, inject_before_detrending=True,
                                          func=func, **{"model":"davenport2014"}, mode="custom")
    #make sure no flares are injected overlapping true flares
    data = flcd.fake_flares
    assert data[(data.istart > 14) & (data.istart < 19)].shape[0] == 0
    #test if all injected event are covered in the merged flares:
    assert data.shape[0] == 10
    assert fflc.gaps == [(0, 1000)]
    assert np.median(fflc.it_med.value) == pytest.approx(500.005274113832/2.)
    assert flcd.detrended_flux.value == pytest.approx(flc.flux/2.)


    # Custom case with detrend_kwargs
    
    def func(flc, kw=0):
        flc.detrended_flux =  flc.flux/2.
        flc.detrended_flux_err =  flc.flux_err/2.
        a = kw + 3
        assert a ==20
        return flc
    
    flc = mock_flc(detrended=True)
    
    flcd, fflc = flc.sample_flare_recovery(iterations=10, inject_before_detrending=True,
                                           func=func, mode="custom", **{"model":"davenport2014"},
                                           detrend_kwargs={"kw":17})
    #make sure no flares are injected overlapping true flares
    data = flcd.fake_flares
    assert data[(data.istart > 14) & (data.istart < 19)].shape[0] == 0
    
    #test if all injected event are covered in the merged flares:
    assert data.shape[0] == 10
    assert fflc.gaps == [(0, 1000)]
    assert float(np.median(fflc.it_med)) == pytest.approx(500.005274113832/2.)
    assert flcd.detrended_flux == pytest.approx(flc.flux/2.)
       
    flcd.flares = flcd.flares.astype(float)
    # Test that the original flare was not changed accidentally
    assert flcd.flares.loc[0,'ed_rec'] == pytest.approx(3455.8875941, rel=1e-4)
    assert flcd.flares['ed_rec_err'][0] < flcd.flares['ed_rec'][0]
    assert flcd.flares['istart'][0] == 15
    assert flcd.flares['istop'][0] == 19
    assert flcd.flares['cstop'][0] == 19
    assert flcd.flares['cstart'][0] == 15
    assert flcd.flares['tstart'][0] == pytest.approx(0.3125, rel=1e-4)
    assert flcd.flares['tstop'][0] == pytest.approx(0.395833, rel=1e-4)
    assert flcd.flares['total_n_valid_data_points'][0] == 1000
    assert flcd.flares['ampl_rec'][0] == pytest.approx(1, rel=1e-3)
    
    # Test that adding another round of injrec will append to the path
    flc = mock_flc(detrended=True)
    flcd, fflc = flc.sample_flare_recovery(iterations=10, inject_before_detrending=False,
                                           save=True, **{"model":"davenport2014"},)
    size = len(flcd.fake_flares)
    
    flcd, fflc = flcd.sample_flare_recovery(iterations=10, inject_before_detrending=False,
                                           save=True, **{"model":"davenport2014"},)
    size2 = len(flcd.fake_flares)
    assert size < size2
    
    path ='10_800000000_inj_after_5.csv'
    saved = pd.read_csv(path)
    assert saved.shape[0] == size2
    
    os.remove(path)


def test_repr():
    pass

def test_getitem():
    pass


def test_invalid_lightcurve():
    """Invalid FlareLightCurves should not be allowed."""
    err_string = ("Input arrays have different lengths."
                  " len(time)=5, len(flux)=4")
    time = np.array([1, 2, 3, 4, 5])
    flux = np.array([1, 2, 3, 4])
    with pytest.raises(ValueError) as err:
        FlareLightCurve(time=time, flux=flux)


def test_find_gaps():
    flux = np.random.rand(1000)
    time = np.linspace(0,30,1000)
    flux[20:200] = np.nan
    time = time[np.where(~np.isnan(flux))]
    flux = flux[np.where(~np.isnan(flux))]
    flc = FlareLightCurve(time=time, flux=flux)

    flc = flc.find_gaps()
    assert flc.gaps == [(0, 20), (20, 820)]

def test_detrend():
    # Test SAVGOL detrending
    
    ampls = [100, 10., 1, .1, .01]
    durs = [1, 2, 3]
    lcs = []
    for ampl in ampls:
        for dur in durs:
            aplc = mock_flc(ampl=ampl, dur=dur)
            daplc = aplc.detrend("savgol")
            lcs.append(daplc)

    for daplc in lcs:
        print("FFFDF", daplc.detrended_flux)
        fff = find_iterative_median(daplc)
        print("DF", fff.detrended_flux)
        shape = fff.flux.value.shape
        assert fff.it_med.value == pytest.approx(500., rel=0.01) #median stays the same roughly
        assert aplc.flux.value.shape[0] == daplc.detrended_flux.value.shape[0] #no NaNs to throw out
        assert daplc.flux.value.max() > daplc.detrended_flux.value.max() # flare sits on a LC part above quiescent level
        assert (aplc.flux_err.value == daplc.detrended_flux_err.value).all() # uncertainties are simply kept
        # Test that shapes of arrays are kept
        for att in ["detrended_flux", "detrended_flux_err",
            "flux_err", "flux", "time", "quality"]:
            assert getattr(fff, att).value.shape == shape
        
    # TEST CUSTOM DETRENDING
    
    # --- create a very minimalistic light curve
    N = int(1e4)
    time = np.linspace(2000,2050,N)
    np.random.seed(200)
    flux = 5e4 + np.random.rand(N) * 35. 
    flux_err = np.random.rand(N) * 35. 
    flc = FlareLightCurve(targetid=10000009, time=time, flux=flux, flux_err=flux_err)

    # --- test a minimum function that fails to create the desired output
    def custom_detrending(flc):
        return flc

    with pytest.raises(AttributeError) as e:
        new_flc = flc.detrend(mode="custom", func=custom_detrending)

    # -- test a minimum function that does the job    
    def custom_detrending(flc):
        flc.detrended_flux = flc.flux
        flc.detrended_flux_err = flc.flux_err
        return flc    
        
    new_flc = flc.detrend(mode="custom", func=custom_detrending)
    assert (new_flc.flux.value == flc.flux.value).all()
    assert (new_flc.flux_err.value == flc.flux_err.value).all()

    # -- test a minimum function that does the job and has kwargs
    def custom_detrending(flc, kw=0):
        flc.detrended_flux = flc.flux
        flc.detrended_flux_err = flc.flux_err
        a = kw + 3 
        assert a == 20
        return flc    
        
    new_flc = flc.detrend(mode="custom", func=custom_detrending, kw=17)
    assert (new_flc.flux.value == flc.flux.value).all()
    assert (new_flc.flux_err.value == flc.flux_err.value).all()

    # --- function should fail if no func is given

    with pytest.raises(ValueError) as e:
        new_flc = flc.detrend(mode="custom")


def test_detrend_fails():
    """If detrend fails, an error is raised with given string."""
    
    # De-trending fails in general when an invalid mode is passed.
    # But also a helpful message is thrown out.
    flc =  mock_flc()
    err_string = ('\nDe-trending mode blaaaah does not exist. Pass "savgol" (for a Savitzky-Golay filter based detrending)'
                       ' or "custom" (to pass a custom detrending function to func=).')
    with pytest.raises(ValueError) as err:
        flc.detrend("blaaaah")
    assert err_string == err.value.args[0]

def test_find_flares():
    """Test that an obvious flare is recovered sufficiently well."""
    flc = mock_flc(detrended=True)
    flc = flc.find_flares()
    #print(flc.flares)
    assert flc.flares.loc[0,'ed_rec'] == pytest.approx(3455.8875941, rel=1e-4)
    assert flc.flares['ed_rec_err'][0] < flc.flares['ed_rec'][0]
    assert flc.flares['istart'][0] == 15
    assert flc.flares['istop'][0] == 19
    assert flc.flares['cstop'][0] == 19
    assert flc.flares['cstart'][0] == 15
    assert flc.flares['tstart'][0] == pytest.approx(0.3125, rel=1e-4)
    assert flc.flares['tstop'][0] == pytest.approx(0.395833, rel=1e-4)
    assert flc.flares['total_n_valid_data_points'][0] == 1000
    assert flc.flares['ampl_rec'][0] == pytest.approx(1, rel=1e-3)
    

def test_inject_fake_flares():
    flc = mock_flc(detrended=True)
    np.random.seed(84712)
    flc = flc.find_gaps()
    fake_flc = flc.inject_fake_flares()
    # make sure you inject only one flare per LC
    assert len(fake_flc.gaps) == fake_flc.fake_flares.shape[0]
    assert (set(fake_flc.fake_flares.columns.values.tolist()) == 
            {'amplitude', 'duration_d', 'ed_inj', 'peak_time'})
    assert fake_flc.detrended_flux_err.value.all() >= 1e-10
    assert fake_flc.detrended_flux.value.all() <= 1.
    assert fake_flc.detrended_flux.value.shape == flc.detrended_flux.value.shape
    flc = mock_flc(detrended=False)
    np.random.seed(84712)
    flc = flc.find_gaps()
    fake_flc = flc.inject_fake_flares(inject_before_detrending=True)

    # make sure you inject only one flare per LC
    assert len(fake_flc.gaps) == fake_flc.fake_flares.shape[0]
    assert (set(fake_flc.fake_flares.columns.values.tolist()) == 
            {'amplitude', 'duration_d', 'ed_inj', 'peak_time'})
    assert fake_flc.flux_err.value.all() >= 1e-10
    assert fake_flc.flux.value.all() <= 1.
    assert fake_flc.flux.value.shape == flc.flux.value.shape

def test_load_injrec_data():
    # Create a minimal empty light curve with an ID
    flcd = FlareLightCurve(targetid="GJ 1243", time=np.linspace(10,1))

    # Path to test file
    path = "altaipony/tests/testfiles/gj1243_injrec.csv"

    # Call the function for the first time
    flcd.load_injrec_data(path)

    # Check if nothing happened to the size
    assert flcd.fake_flares.shape[0] == 1010
    assert flcd.fake_flares.shape[1] == 14 

    # Loading a second time should append the new table
    flcd.load_injrec_data(path)
    
    # Twice as many rows, but same number of columns
    assert flcd.fake_flares.shape[0] == 2020
    assert flcd.fake_flares.shape[1] == 14 

    # We should get a FileNotFoundError when a bad path is passed:
    with pytest.raises(FileNotFoundError) as err:
        flcd.load_injrec_data("badpath")

def test_plot_ed_ratio_heatmap():
    # Create a minimal empty light curve with an ID
    flcd = FlareLightCurve(targetid="GJ 1243", time=np.linspace(10,1))
    
    # Path to test file
    path = "altaipony/tests/testfiles/gj1243_injrec.csv"
    flcd.load_injrec_data(path)
    
    # Test if the function is called properly with default values
    flcd.plot_ed_ratio_heatmap()


def test_plot_recovery_probability_heatmap():
    # Create a minimal empty light curve with an ID
    flcd = FlareLightCurve(targetid="GJ 1243", time=np.linspace(10,1))
    
    # Path to test file
    path = "altaipony/tests/testfiles/gj1243_injrec.csv"
    flcd.load_injrec_data(path)
    
    # Test if the function is called properly with default values
    flcd.plot_recovery_probability_heatmap()


# ------------------------------------------------------------------------------

